using System;
using System.Collections.Generic;

namespace Smdn.Protocols.Imap4 {
  public static class ImapDataConverter {
    public static bool ToMsgAtt(ImapData[] list, out ImapMessageFlags? flags, out string[] extendedFlags, out ImapEnvelope envelope, out DateTime? internalDate, out int? size, out ImapBodyStructure bodyStructure, out string body, out int? uid)
    {
      /*
      msg-att         = "(" (msg-att-dynamic / msg-att-static)
                         *(SP (msg-att-dynamic / msg-att-static)) ")"
      msg-att-dynamic = "FLAGS" SP "(" [flag-fetch *(SP flag-fetch)] ")"
                          ; MAY change for a message
      msg-att-static  = "ENVELOPE" SP envelope / "INTERNALDATE" SP date-time /
                        "RFC822" [".HEADER" / ".TEXT"] SP nstring /
                        "RFC822.SIZE" SP number /
                        "BODY" ["STRUCTURE"] SP body /
                        "BODY" section ["<" number ">"] SP nstring /
                        "UID" SP uniqueid
                          ; MUST NOT change for a message
      */
      flags           = null;
      extendedFlags   = null;
      envelope        = null;
      internalDate    = null;
      // TODO: rfc822header = null;
      // TODO: rfc822text = null;
      size            = null;
      bodyStructure   = null;
      body            = null;
      uid             = null;

      if (list.Length % 2 != 0) // must be 'att SP value'
        return false;

      for (var index = 0; index < list.Length; index++) {
        if (list[index].Format != ImapDataFormat.Text)
          continue;

        switch (list[index].Text.ToUpper()) {
          case "FLAGS":
            ImapMessageFlags f;
            if (!ToFlag(list[index + 1].List, out f, out extendedFlags))
              continue;
            flags = f;
            break;

          case "ENVELOPE":
            if (!ToEnvelope(list[index + 1].List, out envelope))
              continue;
            break;

          case "INTERNALDATE":
            internalDate = ToDateTime(list[index + 1]);
            break;

          case "RFC822.HEADER":
            // TODO: impl
            break;

          case "RFC822.TEXT":
            // TODO: impl
            break;

          case "RFC822.SIZE":
            size = ToNumber(list[index + 1]);
            break;

          case "BODY":
          case "BODYSTRUCTURE":
            if (!ToBody(list[index + 1].List, out bodyStructure))
              continue;
            break;

          case "RFC822":
          case "BODY[]":
            if (list[index + 1].Format != ImapDataFormat.Text)
              continue;
            body = list[index + 1].Text;
            break;

          case "UID":
            uid = ToNumber(list[index + 1]);
            break;
        }
      }

      return true;
    }

    public static bool ToStatusAtt(ImapData[] list, out int? messages, out int? recent, out int? unseen, out int? uidNext, out int? uidValidity)
    {
      /*
      status-att      = "MESSAGES" / "RECENT" / "UIDNEXT" / "UIDVALIDITY" /
                        "UNSEEN"
      status-att-list =  status-att SP number *(SP status-att SP number)
      */
      messages    = null;
      recent      = null;
      unseen      = null;
      uidNext     = null;
      uidValidity = null;

      if (list.Length % 2 != 0) // must be 'status-att SP number'
        return false;

      if (!IsRequiredFormat(list, null, ImapDataFormat.Text))
        return false;

      for (var index = 0; index < list.Length; index += 2) {
        switch (list[index].Text.ToUpper()) {
          case "MESSAGES":
            messages = ToNumber(list[index + 1]); 
            break;

          case "RECENT":
            recent = ToNumber(list[index + 1]); 
            break;

          case "UIDNEXT":
            uidNext = ToNumber(list[index + 1]); 
            break;

          case "UIDVALIDITY":
            uidValidity = ToNumber(list[index + 1]); 
            break;

          case "UNSEEN":
            unseen = ToNumber(list[index + 1]); 
            break;
        }
      }

      return true;
    }

    public static bool ToCapabilityData(ImapData[] list, out ImapCapability capability)
    {
      /*
      capability      = ("AUTH=" auth-type) / atom
                          ; New capabilities MUST begin with "X" or be
                          ; registered with IANA as standard or
                          ; standards-track
      capability-data = "CAPABILITY" *(SP capability) SP "IMAP4rev1"
                        *(SP capability)
                          ; Servers MUST implement the STARTTLS, AUTH=PLAIN,
                          ; and LOGINDISABLED capabilities
                          ; Servers which offer RFC 1730 compatibility MUST
                          ; list "IMAP4" as the first capability.
      */
      capability = null;

      if (!IsRequiredFormat(list, new[] {
        new [] {ImapDataFormat.Text}, // first 'CAPABILITY'
        new [] {ImapDataFormat.Text}, // 'IMAP4rev1'
      }, ImapDataFormat.Text))
        return false;

      if (!IsCaseInsensitiveTextEqualsTo(list[0], "CAPABILITY"))
        return false;

      var knownCaps = ImapKnownCapabilities.None;
      var allCaps = new List<string>();

      for (var index = 1; index < list.Length; index++ ) {
        allCaps.Add(list[index].Text);

        switch (list[index].Text.ToUpper()) {
          case "IMAP4REV1":
            knownCaps |= ImapKnownCapabilities.Imap4Rev1;
            break;

          case "AUTH=KERBEROS_V4":
            knownCaps |= ImapKnownCapabilities.AuthKerberosV4;
            break;

          case "AUTH=CRAM-MD5":
            knownCaps |= ImapKnownCapabilities.AuthCramMd5;
            break;

          case "AUTH=PLAIN":
            knownCaps |= ImapKnownCapabilities.AuthPlain;
            break;

          case "AUTH=LOGIN":
            knownCaps |= ImapKnownCapabilities.AuthLogin;
            break;

          case "IDLE":
            knownCaps |= ImapKnownCapabilities.Idle;
            break;

          case "NAMESPACE":
            knownCaps |= ImapKnownCapabilities.Namespace;
            break;

          case "STARTTLS":
            knownCaps |= ImapKnownCapabilities.StartTls;
            break;

          case "LOGINDISABLED":
            knownCaps |= ImapKnownCapabilities.LoginDisabled;
            break;

          case "CHILDREN":
            knownCaps |= ImapKnownCapabilities.Children;
            break;

          case "XLIST":
            knownCaps |= ImapKnownCapabilities.Xlist;
            break;

          case "XYZZY":
            knownCaps |= ImapKnownCapabilities.Xyzzy;
            break;
        }
      }

      capability = new ImapCapability(knownCaps, allCaps.ToArray());

      return true;
    }

    public static bool ToFlag(ImapData[] list, out ImapMessageFlags flags, out string[] extendedFlags)
    {
      /*
      flag            = "\Answered" / "\Flagged" / "\Deleted" /
                        "\Seen" / "\Draft" / flag-keyword / flag-extension
                          ; Does not include "\Recent"
      flag-extension  = "\" atom
                          ; Future expansion.  Client implementations
                          ; MUST accept flag-extension flags.  Server
                          ; implementations MUST NOT generate
                          ; flag-extension flags except as defined by
                          ; future standard or standards-track
                          ; revisions of this specification.
      flag-fetch      = flag / "\Recent"
      flag-keyword    = atom
      flag-list       = "(" [flag *(SP flag)] ")"
      flag-perm       = flag / "\*"
      */

      flags = ImapMessageFlags.None;
      extendedFlags = new string[] {};

      if (!IsRequiredFormat(list, null, ImapDataFormat.Text))
        return false;

      var ext = new List<string>();

      foreach (var data in list) {
        switch (data.Text) {
          case @"\Seen":
            flags |= ImapMessageFlags.Seen;
            break;
          case @"\Answered":
            flags |= ImapMessageFlags.Answered;
            break;
          case @"\Flagged":
            flags |= ImapMessageFlags.Flagged;
            break;
          case @"\Deleted":
            flags |= ImapMessageFlags.Deleted;
            break;
          case @"\Draft":
            flags |= ImapMessageFlags.Draft;
            break;
          case @"\Recent":
            flags |= ImapMessageFlags.Recent;
            break;

          case @"\*":
            flags |= ImapMessageFlags.AllowedCreateKeywords;
            break;

          default:
            ext.Add(data.Text);
            break;
        }
      }

      extendedFlags = ext.ToArray();

      return true;
    }

    public static bool ToMbxListFlags(ImapData[] list, out ImapMailboxFlags flags, out string[] extendedFlags)
    {
      /*
      mbx-list-flags  = *(mbx-list-oflag SP) mbx-list-sflag
                        *(SP mbx-list-oflag) /
                        mbx-list-oflag *(SP mbx-list-oflag)
      mbx-list-oflag  = "\Noinferiors" / flag-extension
                          ; Other flags; multiple possible per LIST response
      mbx-list-sflag  = "\Noselect" / "\Marked" / "\Unmarked"
                          ; Selectability flags; only one per LIST response
      */

      flags = ImapMailboxFlags.None;
      extendedFlags = new string[] {};

      if (!IsRequiredFormat(list, null, ImapDataFormat.Text))
        return false;

      var ext = new List<string>();

      foreach (var data in list) {
        switch (data.Text) {
          case @"\Marked":
            flags |= ImapMailboxFlags.Marked;
            break;
          case @"\Unmarked":
            flags |= ImapMailboxFlags.UnMarked;
            break;
          case @"\Noinferiors":
            flags |= ImapMailboxFlags.NoInferiors;
            break;
          case @"\Noselect":
            flags |= ImapMailboxFlags.NoSelect;
            break;

          /*
           * RFC 3348 The Internet Message Action Protocol (IMAP4) Child Mailbox Extension
           * http://tools.ietf.org/html/rfc3348
           */
          case @"\HasChildren":
            flags |= ImapMailboxFlags.HasChildren;
            break;
          case @"\HasNoChildren":
            flags |= ImapMailboxFlags.HasNoChildren;
            break;

          default:
            ext.Add(data.Text);
            break;
        }
      }

      extendedFlags = ext.ToArray();

      return true;
    }

    public static bool ToEnvelope(ImapData[] list, out ImapEnvelope envelope)
    {
      /*
      envelope        = "(" env-date SP env-subject SP env-from SP
                        env-sender SP env-reply-to SP env-to SP env-cc SP
                        env-bcc SP env-in-reply-to SP env-message-id ")"
      env-bcc         = "(" 1*address ")" / nil
      env-cc          = "(" 1*address ")" / nil
      env-date        = nstring
      env-from        = "(" 1*address ")" / nil
      env-in-reply-to = nstring
      env-message-id  = nstring
      env-reply-to    = "(" 1*address ")" / nil
      env-sender      = "(" 1*address ")" / nil
      env-subject     = nstring
      env-to          = "(" 1*address ")" / nil
      */
      envelope = null;

      if (list.Length != 10)
        return false;

      var index = 0;
      var envData = new object[list.Length];

      foreach (var format in new[] {
        new[] { ImapDataFormat.Text, ImapDataFormat.Nil }, // [0] env-date
        new[] { ImapDataFormat.Text, ImapDataFormat.Nil }, // [1] env-subject
        new[] { ImapDataFormat.List, ImapDataFormat.Nil }, // [2] env-from
        new[] { ImapDataFormat.List, ImapDataFormat.Nil }, // [3] env-sender
        new[] { ImapDataFormat.List, ImapDataFormat.Nil }, // [4] env-reply-to
        new[] { ImapDataFormat.List, ImapDataFormat.Nil }, // [5] env-to
        new[] { ImapDataFormat.List, ImapDataFormat.Nil }, // [6] env-cc
        new[] { ImapDataFormat.List, ImapDataFormat.Nil }, // [7] env-bcc
        new[] { ImapDataFormat.Text, ImapDataFormat.Nil }, // [8] env-in-reply-to
        new[] { ImapDataFormat.Text, ImapDataFormat.Nil }, // [9] env-message-id
      }) {
        var converted = false;

        foreach (var f in format) {
          if (list[index].Format == f) {
            converted = true;
            switch (f) {
              case ImapDataFormat.Text:
                envData[index] = list[index].Text;
                break;

              case ImapDataFormat.List:
                if (list[index].List.Length != 1)
                  return false;
                if (list[index].List[0].Format != ImapDataFormat.List)
                  return false;

                ImapAddress address;
                if (!ToAddress(list[index].List[0].List, out address))
                  return false;
                envData[index] = address;
                break;

              case ImapDataFormat.Nil:
                envData[index] = null;
                break;
            }

            break; // converted
          }
        }

        if (!converted)
          return false;

        index++;
      }

      envelope = new ImapEnvelope(envData[0] as string, // env-date
                                  envData[1] as string, // env-subject
                                  envData[2] as ImapAddress, // env-from
                                  envData[3] as ImapAddress, // env-sender
                                  envData[4] as ImapAddress, // env-reply-to
                                  envData[5] as ImapAddress, // env-to
                                  envData[6] as ImapAddress, // env-cc
                                  envData[7] as ImapAddress, // env-bcc
                                  envData[8] as string, // env-in-reply-to
                                  envData[9] as string // env-message-id
                                  );
      return true;
    }

    public static bool ToAddress(ImapData[] list, out ImapAddress address)
    {
      /*
      address         = "(" addr-name SP addr-adl SP addr-mailbox SP
                        addr-host ")"
      addr-adl        = nstring
                          ; Holds route from [RFC-2822] route-addr if
                          ; non-NIL
      addr-host       = nstring
                          ; NIL indicates [RFC-2822] group syntax.
                          ; Otherwise, holds [RFC-2822] domain name
      addr-mailbox    = nstring
                          ; NIL indicates end of [RFC-2822] group; if
                          ; non-NIL and addr-host is NIL, holds
                          ; [RFC-2822] group name.
                          ; Otherwise, holds [RFC-2822] local-part
                          ; after removing [RFC-2822] quoting
      addr-name       = nstring
                          ; If non-NIL, holds phrase from [RFC-2822]
                          ; mailbox after removing [RFC-2822] quoting
      */
      address = null;

      if (!IsRequiredFormat(list, new[] {
        new [] {ImapDataFormat.Text, ImapDataFormat.Nil},
        new [] {ImapDataFormat.Text, ImapDataFormat.Nil},
        new [] {ImapDataFormat.Text, ImapDataFormat.Nil},
        new [] {ImapDataFormat.Text, ImapDataFormat.Nil},
      }, ImapDataFormat.None))
        return false;

      address = new ImapAddress(ToNString(list[0]),
                                ToNString(list[1]),
                                ToNString(list[2]),
                                ToNString(list[3]));
      return true;
    }

    public static bool ToBody(ImapData[] list, out ImapBodyStructure bodyStructure)
    {
      /*
      body            = "(" (body-type-1part / body-type-mpart) ")"
      body-extension  = nstring / number /
                         "(" body-extension *(SP body-extension) ")"
                          ; Future expansion.  Client implementations
                          ; MUST accept body-extension fields.  Server
                          ; implementations MUST NOT generate
                          ; body-extension fields except as defined by
                          ; future standard or standards-track
                          ; revisions of this specification.
      body-ext-1part  = body-fld-md5 [SP body-fld-dsp [SP body-fld-lang
                        [SP body-fld-loc *(SP body-extension)]]]
                          ; MUST NOT be returned on non-extensible
                          ; "BODY" fetch
      body-ext-mpart  = body-fld-param [SP body-fld-dsp [SP body-fld-lang
                        [SP body-fld-loc *(SP body-extension)]]]
                          ; MUST NOT be returned on non-extensible
                          ; "BODY" fetch
      body-fields     = body-fld-param SP body-fld-id SP body-fld-desc SP
                        body-fld-enc SP body-fld-octets
      body-fld-desc   = nstring
      body-fld-dsp    = "(" string SP body-fld-param ")" / nil
      body-fld-enc    = (DQUOTE ("7BIT" / "8BIT" / "BINARY" / "BASE64"/
                        "QUOTED-PRINTABLE") DQUOTE) / string
      body-fld-id     = nstring
      body-fld-lang   = nstring / "(" string *(SP string) ")"
      body-fld-loc    = nstring
      body-fld-lines  = number
      body-fld-md5    = nstring
      body-fld-octets = number
      body-fld-param  = "(" string SP string *(SP string SP string) ")" / nil
      body-type-1part = (body-type-basic / body-type-msg / body-type-text)
                        [SP body-ext-1part]
      body-type-basic = media-basic SP body-fields
                          ; MESSAGE subtype MUST NOT be "RFC822"
      body-type-mpart = 1*body SP media-subtype
                        [SP body-ext-mpart]
      body-type-msg   = media-message SP body-fields SP envelope
                        SP body SP body-fld-lines
      body-type-text  = media-text SP body-fields SP body-fld-lines

      media-basic     = ((DQUOTE ("APPLICATION" / "AUDIO" / "IMAGE" /
                        "MESSAGE" / "VIDEO") DQUOTE) / string) SP
                        media-subtype
                          ; Defined in [MIME-IMT]
      media-message   = DQUOTE "MESSAGE" DQUOTE SP DQUOTE "RFC822" DQUOTE
                          ; Defined in [MIME-IMT]
      media-subtype   = string
                          ; Defined in [MIME-IMT]
      media-text      = DQUOTE "TEXT" DQUOTE SP media-subtype
                          ; Defined in [MIME-IMT]
       */
      bodyStructure = null;

      if (list.Length < 1)
        return false;

      if (list[0].Format == ImapDataFormat.Text)
        return ToNonMultipartBody(list, out bodyStructure);
      else if (list[0].Format == ImapDataFormat.List)
        return ToMultipartBody(list, out bodyStructure);
      else
        return false;
    }

    private static bool ToNonMultipartBody(ImapData[] list, out ImapBodyStructure bodyStructure)
    {
      bodyStructure = null;

      // basic fields
      //   type
      //     [0] type:        media-basic / media-text / media-message
      //     [1] subtype:     media-subtype
      //   body-fields
      //     [2] parameters:  body-fld-param
      //     [3] id:          body-fld-id
      //     [4] description: body-fld-desc
      //     [5] encoding:    body-fld-enc
      //     [6] size:        body-fld-octets
      //   <body-type-msg>       <body-type-text>       <body-type-basic>
      //     [7] envelope          [7] body-fld-lines     [7] (body-ext-1part)
      //     [8] body              [8] (body-ext-1part)
      //     [9] body-fld-lines
      //     [10] (body-ext-1part)
      if (!IsRequiredFormat(list, new[] {
        new [] {ImapDataFormat.Text},
        new [] {ImapDataFormat.Text},
        new [] {ImapDataFormat.List, ImapDataFormat.Nil},
        new [] {ImapDataFormat.Text, ImapDataFormat.Nil},
        new [] {ImapDataFormat.Text, ImapDataFormat.Nil},
        new [] {ImapDataFormat.Text},
        new [] {ImapDataFormat.Text},
      }, ImapDataFormat.None))
        return false;

      var basicFields = new object[7];

      for (var i = 0; i < basicFields.Length; i++) {
        basicFields[i] = null;
      }

      // type
      string bodyType;

      if (IsCaseInsensitiveTextEqualsTo(list[0], "MESSAGE")) {
        if (IsCaseInsensitiveTextEqualsTo(list[1], "RFC822"))
          // body-type-msg
          bodyType = "msg";
        else
          // body-type-basic
          bodyType = "basic";
      }
      else if (IsCaseInsensitiveTextEqualsTo(list[0], "TEXT")) {
         // body-type-text
        bodyType = "text";
      }
      else {
        // body-type-basic
        bodyType = "basic";
      }

      basicFields[0] = list[0].Text; // [0] type
      basicFields[1] = list[1].Text; // [1] subtype

      // body-fields

      // [2] parameters
      if (list[2].Format == ImapDataFormat.Nil) {
        basicFields[2] = null;
      }
      else if (list[2].Format == ImapDataFormat.List) {
        basicFields[2] = ToAttributeValuePairs(list[2].List);
      }
      else {
        return false;
      }

      basicFields[3] = ToNString(list[3]); // [3] id
      basicFields[4] = ToNString(list[4]); // [4] description
      basicFields[5] = ToNString(list[5]); // [5] encoding
      basicFields[6] = ToNumber(list[6]);  // [6] size

      ImapEnvelope encapsulatedEnvelope = null;
      ImapBodyStructure encapsulatedBodyStructure = null;

      if (bodyType == "msg") {
        // [7] envelope
        if (!ToEnvelope(list[7].List, out encapsulatedEnvelope))
          return false;

        // [8] body
        if (!ToBody(list[8].List, out encapsulatedBodyStructure))
          return false;
      }

      int lines = 0;

      if (bodyType == "msg")
        lines = ToNumber(list[9]) ?? 0; // [9] body-fld-lines
      else if (bodyType == "text")
        lines = ToNumber(list[7]) ?? 0; // [7] body-fld-lines

      // TODO: parse extension data
      //    body MD5
      //    body disposition
      //    body language
      //    body location

      bodyStructure = new ImapBodyStructure(null,
                                            basicFields[0] as string,
                                            basicFields[1] as string,
                                            basicFields[2] as Dictionary<string, string>,
                                            basicFields[3] as string,
                                            basicFields[4] as string,
                                            basicFields[5] as string,
                                            (basicFields[6] == null) ? 0 : (int)basicFields[6],
                                            encapsulatedEnvelope,
                                            encapsulatedBodyStructure,
                                            lines);
      return true;
    }

    private static bool ToMultipartBody(ImapData[] list, out ImapBodyStructure bodyStructure)
    {
      bodyStructure = null;

      if (list.Length < 2)
        return false;

      var nested = new List<ImapBodyStructure>();
      var subType = string.Empty;

      foreach (var data in list) {
        if (data.Format == ImapDataFormat.List) {
          // body(nested)
          ImapBodyStructure nestedBodyStructure;
          if (ToBody(data.List, out nestedBodyStructure))
            nested.Add(nestedBodyStructure);
        }
        else if (data.Format == ImapDataFormat.Text) {
          // media-subtype
          subType = data.Text;
        }
        else {
          return false;
        }
      }

      // TODO: parse extension data
      //    body parameter parenthesized list
      //    body disposition
      //    body language
      //    body location

      bodyStructure = new Smdn.Protocols.Imap4.ImapBodyStructure(nested.ToArray(),
                                                                 string.Empty,
                                                                 subType,
                                                                 null,
                                                                 string.Empty,
                                                                 string.Empty,
                                                                 string.Empty,
                                                                 0,
                                                                 null,
                                                                 null,
                                                                 0);

      return true;
    }


    public static Dictionary<string, string> ToAttributeValuePairs(ImapData[] list)
    {
      if (list.Length % 2 != 0) // must be list of attribute/value pairs
        return null;

      if (!IsRequiredFormat(list, null, ImapDataFormat.Text))
        return null;

      var pairs = new Dictionary<string, string>();

      for (var param = 0; param < list.Length; param += 2) {
        pairs.Add(list[param].Text, list[param + 1].Text);
      }

      return pairs;
    }

    public static string[] ToTextList(ImapData[] list)
    {
      if (!IsRequiredFormat(list, null, ImapDataFormat.Text))
        return null;

      var testList = new List<string>();

      foreach (var text in list) {
        testList.Add(text.Text);
      }

      return testList.ToArray();
    }

    public static DateTime? ToDateTime(ImapData textData)
    {
      if (textData.Format == ImapDataFormat.Text) {
        DateTime dt;
        if (DateTime.TryParse(textData.Text, out dt))
          return dt;
        else
          return null;
      }
      else {
        return null;
      }
    }

    public static string ToNString(ImapData textData)
    {
      if (textData.Format == ImapDataFormat.Text)
        return textData.Text;
      else if (textData.Format == ImapDataFormat.Nil)
        return null;
      else
        return null;
    }

    public static int? ToNumber(ImapData textData)
    {
      int number;

      if (int.TryParse(textData.Text, out number))
        return number;
      else
        return null;
    }

    /*
     * utility methods
     */

    public static bool IsRequiredFormat(ImapData[] list, ImapDataFormat[][] listFormats, ImapDataFormat allowedForFollowing)
    {
      if (listFormats != null) {
        // check whether list is required format
        if (list.Length < listFormats.Length)
          return false;

        for (var index = 0; index < listFormats.Length; index++) {
          var matched = false;
          foreach (var f in listFormats[index]) {
            if (list[index].Format == f) {
              matched = true;
              break;
            }
          }
          if (!matched)
            return false;
        }
      }

      if (allowedForFollowing == ImapDataFormat.None)
        return true;

      // check whether following data in list is allowed format
      var length = (listFormats == null) ? 0 : listFormats.Length;

      for (var index = length; index < list.Length; index++ ) {
        if (list[index].Format != allowedForFollowing)
          return false;
      }

      return true;
    }

    public static bool IsCaseInsensitiveTextEqualsTo(ImapData textData, string text)
    {
      return textData.Text.Equals(text, StringComparison.InvariantCultureIgnoreCase);
    }
  }
}