using System;
using System.Collections.Generic;

namespace Smdn.Protocols.Imap4.Protocol {
  // format converting methods for ImapDataResponse
  internal static class ImapDataResponseConverter {
    /*
     * 6. Client Commands
     */

    // 6.3.1. SELECT Command
    //   Responses:  REQUIRED untagged responses: FLAGS, EXISTS, RECENT
    //               REQUIRED OK untagged responses:  UNSEEN,  PERMANENTFLAGS,
    //               UIDNEXT, UIDVALIDITY
    // 6.3.2. EXAMINE Command
    //   Responses:  REQUIRED untagged responses: FLAGS, EXISTS, RECENT
    //               REQUIRED OK untagged responses:  UNSEEN,  PERMANENTFLAGS,
    //               UIDNEXT, UIDVALIDITY
    public static bool FromSelectExpungeCommandResponse(ImapDataResponse response, out ImapMessageFlags? flags, out string[] extendedFlags, out int? exists, out int? recent)
    {
      flags         = null;
      extendedFlags = null;
      exists        = null;
      recent        = null;

      // FLAGS
      ImapMessageFlags respFlags;
      string[] respExtendedFlags;

      if (FromFlagsResponse(response, out respFlags, out respExtendedFlags)) {
        flags = respFlags;
        extendedFlags = respExtendedFlags;
        return true;
      }

      // EXISTS
      int respExists;

      if (FromExistsResponse(response, out respExists)) {
        exists = respExists;
        return true;
      }

      // RECENT
      int respRecent;

      if (FromRecentResponse(response, out respRecent)) {
        recent = respRecent;
        return true;
      }

      return false;
    }

    /*
     * RFC 2177 IMAP4 IDLE command
     * http://tools.ietf.org/html/rfc2177
     */

    // 3. Specification
    //   IDLE Command
    //   The IDLE command remains active until the client
    //   responds to the continuation, and as long as an IDLE command is
    //   active, the server is now free to send untagged EXISTS, EXPUNGE, and
    //   other messages at any time
    public static bool FromIdleCommandResponse(ImapDataResponse response, out int? exists, out int? expunge)
    {
      exists = null;
      expunge = null;

      // EXISTS
      int respExists;

      if (FromExistsResponse(response, out respExists)) {
        exists = respExists;
        return true;
      }

      // EXPUNGE
      int respExpunge;

      if (FromExpungeResponse(response, out respExpunge)) {
        expunge = respExpunge;
        return true;
      }

      return false;
    }

    /*
     * 7.2. Server Responses - Server and Mailbox Status
     */

    // 7.2.1 CAPABILITY Response
    //   Contents:   capability listing
    //   Example:    S: * CAPABILITY IMAP4rev1 STARTTLS AUTH=GSSAPI XPIG-LATIN
    public static bool FromCapabilityResponse(ImapDataResponse response, out ImapCapability capability)
    {
      // 1. check response format
      // 2. parse response
      // 3. set result
      return ImapDataConverter.ToCapabilityData(response.Data, out capability);
    }

    // 7.2.2 LIST Response
    //   Contents:   name attributes
    //               hierarchy delimiter
    //               name
    //   Example:    S: * LIST (\Noselect) "/" ~/Mail/foo
    // 7.2.3 LSUB Response
    //   Contents:   name attributes
    //               hierarchy delimiter
    //               name
    //   Example:    S: * LSUB () "." #news.comp.mail.misc
    public static bool FromListLsubResponse(ImapDataResponse response, out ImapMailboxFlags nameAttributes, out string[] extendedNameAttributes, out string hierarchyDelimiter, out string name)
    {
      nameAttributes = ImapMailboxFlags.None;
      extendedNameAttributes = new string[] {};
      hierarchyDelimiter = null;
      name = string.Empty;
          
      // 1. check response format
      if (!ImapDataConverter.IsRequiredFormat(response.Data, new[] {
        new [] {ImapDataFormat.Text},
        new [] {ImapDataFormat.List},
        new [] {ImapDataFormat.Text, ImapDataFormat.Nil},
        new [] {ImapDataFormat.Text},
      }, ImapDataFormat.None))
        return false;
        
      if (!(ImapDataConverter.IsCaseInsensitiveTextEqualsTo(response.Data[0], "LIST") ||
            ImapDataConverter.IsCaseInsensitiveTextEqualsTo(response.Data[0], "LSUB")))
        return false;

      // 2. parse response
      // 3. set result
      hierarchyDelimiter = ImapDataConverter.ToNString(response.Data[2]);
      name = response.Data[3].Text;

      return ImapDataConverter.ToMbxListFlags(response.Data[1].List, out nameAttributes, out extendedNameAttributes);
    }

    // 7.2.4 STATUS Response
    //   Contents:   name
    //               status parenthesized list
    //   Example:    S: * STATUS blurdybloop (MESSAGES 231 UIDNEXT 44292)
    public static bool FromStatusResponse(ImapDataResponse response, out string name, out int? messages, out int? recent, out int? unseen, out int? uidNext, out int? uidValidity)
    {
      /*
      status          = "STATUS" SP mailbox SP
                        "(" status-att *(SP status-att) ")"
      */

      name        = string.Empty;
      messages    = null;
      recent      = null;
      unseen      = null;
      uidNext     = null;
      uidValidity = null;

      // 1. check response format
      if (!ImapDataConverter.IsRequiredFormat(response.Data, new[] {
        new [] {ImapDataFormat.Text},
        new [] {ImapDataFormat.Text},
        new [] {ImapDataFormat.List},
      }, ImapDataFormat.Text))
        return false;

      if (!ImapDataConverter.IsCaseInsensitiveTextEqualsTo(response.Data[0], "STATUS"))
        return false;

      // 2. parse response
      // 3. set result
      name = response.Data[1].Text;

      return ImapDataConverter.ToStatusAtt(response.Data[2].List, out messages, out recent, out unseen, out uidNext, out uidValidity);
    }

    // 7.2.5 SEARCH Response
    //   Contents:   zero or more numbers
    //   Example:    S: * SEARCH 2 3 6
    public static bool FromSearchResponse(ImapDataResponse response, out int[] messages)
    {
      messages = new int[] {};

      // 1. check response format
      if (!ImapDataConverter.IsRequiredFormat(response.Data, new[] {
        new [] {ImapDataFormat.Text},
      }, ImapDataFormat.Text))
        return false;

      if (!ImapDataConverter.IsCaseInsensitiveTextEqualsTo(response.Data[0], "SEARCH"))
        return false;

      // 2. parse response
      var msgs = new List<int>();

      for (var index = 1; index < response.Data.Length; index++) {
        int? msg = ImapDataConverter.ToNumber(response.Data[index]);

        if (msg != null)
          msgs.Add((int)msg);
      }

      // 3. set result
      messages = msgs.ToArray();

      return true;
    }

    // 7.2.6 FLAGS Response
    //   Contents:   flag parenthesized list
    //   Example:    S: * FLAGS (\Answered \Flagged \Deleted \Seen \Draft)
    public static bool FromFlagsResponse(ImapDataResponse response, out ImapMessageFlags flags, out string[] extendedFlags)
    {
      flags = ImapMessageFlags.None;
      extendedFlags = new string[] {};

      // 1. check response format
      if (!ImapDataConverter.IsRequiredFormat(response.Data, new[] {
        new [] {ImapDataFormat.Text},
        new [] {ImapDataFormat.List},
      }, ImapDataFormat.None))
        return false;

      if (!ImapDataConverter.IsCaseInsensitiveTextEqualsTo(response.Data[0], "FLAGS"))
        return false;

      // 2. parse response
      // 3. set result
      return ImapDataConverter.ToFlag(response.Data[1].List, out flags, out extendedFlags);
    }

    /*
     * 7.3. Server Responses - Mailbox Size
     */

    // 7.3.1. EXISTS Response
    //   Contents:   none
    //   Example:    S: * 23 EXISTS
    public static bool FromExistsResponse(ImapDataResponse response, out int exists)
    {
      return FromSizeOrSequenceNumberResponse(response, "EXISTS", out exists);
    }

    // 7.3.2. RECENT Response
    //   Contents:   none
    //   Example:    S: * 5 RECENT
    public static bool FromRecentResponse(ImapDataResponse response, out int recent)
    {
      return FromSizeOrSequenceNumberResponse(response, "RECENT", out recent);
    }

    /*
     * 7.4. Server Responses - Message Status
     */

    // 7.4.1. EXPUNGE Response
    //   Contents:   none
    //   Example:    S: * 44 EXPUNGE
    public static bool FromExpungeResponse(ImapDataResponse response, out int expunge)
    {
      return FromSizeOrSequenceNumberResponse(response, "EXPUNGE", out expunge);
    }

    // 7.4.2. FETCH Response
    //   Contents:   message data
    //   Example:    S: * 23 FETCH (FLAGS (\Seen) RFC822.SIZE 44827)
    public static bool FromFetchResponse(ImapDataResponse response, out ImapMessage messageData)
    {
      /*
      message-data    = nz-number SP ("EXPUNGE" / ("FETCH" SP msg-att))
       */

      messageData = null;

      // 1. check response format
      if (!ImapDataConverter.IsRequiredFormat(response.Data, new[] {
        new [] {ImapDataFormat.Text},
        new [] {ImapDataFormat.Text},
        new [] {ImapDataFormat.List},
      }, ImapDataFormat.None))
        return false;

      // get sequence number
      int sequence;

      if (!FromSizeOrSequenceNumberResponse(response, "FETCH", out sequence))
        return false;

      // 2. parse response
      ImapMessageFlags?            flags           = null;
      string[]             extendedFlags   = null;
      ImapEnvelope          envelope        = null;
      DateTime?             internalDate    = null;
      // TODO: rfc822header = null;
      // TODO: rfc822text = null;
      int?                  size            = null;
      ImapBodyStructure     bodyStructure   = null;
      string               body            = null;
      int?                  uid             = null;

      if (!ImapDataConverter.ToMsgAtt(response.Data[2].List,
                    out flags,
                    out extendedFlags,
                    out envelope,
                    out internalDate,
                    out size,
                    out bodyStructure,
                    out body,
                    out uid))
        return false;

      // 3. set result
      messageData = new ImapMessage(sequence);

      if (flags != null)          messageData.Flags = (ImapMessageFlags)flags;
      if (extendedFlags != null)  messageData.ExtendedFlags = extendedFlags;
      if (envelope != null)       messageData.Envelope = envelope;
      if (internalDate != null)   messageData.InternalDate = (DateTime)internalDate;
      if (size != null)           messageData.BodySize = (int)size;
      if (bodyStructure != null)  messageData.BodyStructure = bodyStructure;
      if (body != null)           messageData.Body = body;
      if (uid != null)            messageData.Uid = (int)uid;

      return true;
    }

    //   Example:    S: * 23 EXISTS
    //   Example:    S: * 5 RECENT
    //   Example:    S: * 44 EXPUNGE
    //   Example:    S: * 23 FETCH (FLAGS (\Seen) RFC822.SIZE 44827)
    private static bool FromSizeOrSequenceNumberResponse(ImapDataResponse response, string responseType, out int number)
    {
      number = 0;

      // 1. check response format
      if (!ImapDataConverter.IsRequiredFormat(response.Data, new[] {
        new [] {ImapDataFormat.Text},
        new [] {ImapDataFormat.Text},
      }, ImapDataFormat.None))
        return false;

      if (!ImapDataConverter.IsCaseInsensitiveTextEqualsTo(response.Data[1], responseType))
        return false;

      // 2. parse response
      // 3. set result
      if (!int.TryParse(response.Data[0].Text, out number))
        return false;

      return true;
    }

    /*
     * RFC 2342 IMAP4 Namespace
     * http://tools.ietf.org/html/rfc2342
     */

    // 5. NAMESPACE Command
    //   Response:  an untagged NAMESPACE response that contains the prefix
    //                and hierarchy delimiter to the server's Personal
    //                Namespace(s), Other Users' Namespace(s), and Shared
    //                Namespace(s) that the server wishes to expose. The
    //                response will contain a NIL for any namespace class
    //                that is not available. Namespace_Response_Extensions
    //                MAY be included in the response.
    //                Namespace_Response_Extensions which are not on the IETF
    //                standards track, MUST be prefixed with an "X-".
    //   Example 5.4:
    //   S: * NAMESPACE (("" "/")) (("~" "/")) (("#shared/" "/")
    //      ("#public/" "/")("#ftp/" "/")("#news." "."))
    public static bool FromNamespaceResponse(ImapDataResponse response, out ImapNamespace ns)
    {
      ns = null;

      // 1. check response format
      if (!ImapDataConverter.IsRequiredFormat(response.Data, new[] {
        new [] {ImapDataFormat.Text},
        new [] {ImapDataFormat.List, ImapDataFormat.Nil},
        new [] {ImapDataFormat.List, ImapDataFormat.Nil},
        new [] {ImapDataFormat.List, ImapDataFormat.Nil},
      }, ImapDataFormat.None))
        return false;

      if (!ImapDataConverter.IsCaseInsensitiveTextEqualsTo(response.Data[0], "NAMESPACE"))
        return false;

      // 2. parse response
      var nsgroups = new ImapNamespace.NamespaceGroup[3];

      for (int index = 0; index < 3; index++) {
        if (response.Data[index + 1].Format == ImapDataFormat.Nil) {
          // TODO: 名前空間がNILな場合の考慮
          nsgroups[index] = new ImapNamespace.NamespaceGroup(string.Empty, string.Empty, string.Empty);
        }
        else {
          foreach (var nsdata in response.Data[index + 1].List) {
            // TODO: 複数の名前空間が定義されている場合の考慮
            if (3 == nsdata.List.Length) {
              nsgroups[index] = new ImapNamespace.NamespaceGroup(nsdata.List[0].Text, nsdata.List[1].Text, nsdata.List[2].Text);
            }
            else if (2 == nsdata.List.Length) {
              nsgroups[index] = new ImapNamespace.NamespaceGroup(nsdata.List[0].Text, nsdata.List[1].Text, string.Empty);
            }
            else {
              return false;
            }
          }
        }
      }

      // 3. set result
      ns = new ImapNamespace(nsgroups[0], nsgroups[1], nsgroups[2]);

      return true;
    }
  }
}
