using System;
using System.Threading;

using Smdn.Protocols.Imap4.Transaction.Transactions;
using Smdn.Protocols.Imap4.Protocol;

namespace Smdn.Protocols.Imap4.Transaction {
  // Layers:
  //   Client.ImapClient
  //   Client.ImapClientAgent
  //   Session.ImapSessionAgent
  //   Session.ImapSession
  //   Transaction.ImapTransactionAgent
  // * Transaction.ImapTransaction/Transaction.ImapRequest/Transaction.ImapResponse
  //   Protocol.ImapConnection/Protocol.ImapProtocolCommand/Protocol.ImapProtocolResponse
  //   (IMAP4)


  public abstract class ImapTransaction {
    protected delegate void ProcessTransactionDelegate();

    public ImapConnection Connection {
      get { return connection; }
    }

    public ImapRequest Request {
      get { return request; }
    }

    public ImapResponse Response {
      get { return response; }
      protected internal set { response = value; }
    }

    internal int SequenceId {
      get { return sequenceId; }
    }

    protected ImapTransaction(ImapConnection connection, ImapRequest request)
    {
      this.connection = connection;
      this.request = request;
      this.response = null;
    }

    public override string ToString()
    {
      return string.Format("{{Connection={0}, Request={1}, Response={2}}}", connection, request, response);
    }

    public static ImapTransaction Create(ImapConnection connection, ImapRequest request)
    {
      var sequenceId = Interlocked.Increment(ref sequenceIdCount);

      if (sequenceIdCount == int.MaxValue)
        Interlocked.Exchange(ref sequenceIdCount, 0);

      ImapTransaction transaction;

      switch (request.Method) {
        /*
         * transaction for initial server greeting
         */
        case ImapTransactionMethod.Greeting:
          transaction = new GreetingTransaction(connection, request);
          break;

        /*
         * transactions in any state
         */
        case ImapTransactionMethod.Capability:
          transaction = new CapabilityTransaction(connection, request);
          break;

        case ImapTransactionMethod.NoOp:
          transaction = new NoOpTransaction(connection, request);
          break;

        case ImapTransactionMethod.Logout:
          transaction = new LogoutTransaction(connection, request);
          break;

        /*
         * transactions in non authenticated state
         */
        //case ImapTransactionMethod.StartTls:
        //  break;

        case ImapTransactionMethod.Authenticate:
          transaction = new AuthenticateTransaction(connection, request);
          break;

        case ImapTransactionMethod.Login:
          transaction = new LoginTransaction(connection, request);
          break;

        /*
         * transactions in authenticated state
         */
        case ImapTransactionMethod.Select:
          transaction = new SelectTransaction(connection, request);
          break;

        case ImapTransactionMethod.Examine:
          transaction = new ExamineTransaction(connection, request);
          break;

        //case ImapTransactionMethod.Create:
        //  break;
        //case ImapTransactionMethod.Delete:
        //  break;
        //case ImapTransactionMethod.Rename:
        //  break;
        //case ImapTransactionMethod.Subscribe:
        //  break;
        //case ImapTransactionMethod.Unsubscribe:
        //  break;

        case ImapTransactionMethod.List:
          transaction = new ListTransaction(connection, request);
          break;

        case ImapTransactionMethod.Lsub:
          transaction = new LsubTransaction(connection, request);
          break;

        case ImapTransactionMethod.Status:
          transaction = new StatusTransaction(connection, request);
          break;

        //case ImapTransactionMethod.Append:
        //  break;

        /*
         * transactions in selected state
         */

        case ImapTransactionMethod.Check:
          transaction = new CheckTransaction(connection, request);
          break;

        case ImapTransactionMethod.Close:
          transaction = new CloseTransaction(connection, request);
          break;

        //case ImapTransactionMethod.Expunge:
        //  break;

        case ImapTransactionMethod.Search:
          transaction = new SearchTransaction(connection, request);
          break;

        case ImapTransactionMethod.Fetch:
          transaction = new FetchTransaction(connection, request);
          break;

        //case ImapTransactionMethod.Store:
        //  break;
        //case ImapTransactionMethod.Copy:
        //  break;

        /*
         * transactions in authenticated/selected state
         */
        case ImapTransactionMethod.Idle:
          transaction = new IdleTransaction(connection, request);
          break;

        case ImapTransactionMethod.IdleDone:
          transaction = new IdleDoneTransaction(connection, request);
          break;

        case ImapTransactionMethod.Namespace:
          transaction = new NamespaceTransaction(connection, request);
          break;

        /*
         * composite transactions
         */
        /*
        case ImapTransactionMethod.AuthAndCapability:
          transaction = new CompositeTransaction(connection, request, new CompositeTransaction.InnerTransaction[] {
            new CompositeTransaction.InnerTransaction(
              Create(connection, new ImapRequest(ImapTransactionMethod.Capability,   request.Timeout, request.Parameters)),
              true
            ),
            new CompositeTransaction.InnerTransaction(
              Create(connection, new ImapRequest(ImapTransactionMethod.Authenticate, request.Timeout, request.Parameters)),
              false
            ),
            new CompositeTransaction.InnerTransaction(
              Create(connection, new ImapRequest(ImapTransactionMethod.Login,        request.Timeout, request.Parameters)),
              true
            ),
            new CompositeTransaction.InnerTransaction(
              Create(connection, new ImapRequest(ImapTransactionMethod.Namespace,    request.Timeout, request.Parameters)),
              true
            ),
          });
          break;
        */

        /*
         * other transactions
         */
        case ImapTransactionMethod.Xyzzy:
          transaction = new XyzzyTransaction(connection, request);
          break;

        default:
          throw new InvalidOperationException("invalid request");
      }

      transaction.ResetInternal();
      transaction.sequenceId = sequenceId;

      return transaction;
    }

    public virtual void Abort()
    {
      // TODO: 中断前に受信してキューに残っているレスポンスの考慮
      lock (this) {
        Transit(null);

        response = new ImapResponse(ImapTransactionStatus.RequestTimeout, "request timed out");
      }
    }

    protected abstract ProcessTransactionDelegate Reset();

    internal void ResetInternal()
    {
      currentProcess = Reset();
    }

    public void Process()
    {
      // トランザクションを処理する
      try {
        if (currentProcess != null)
          currentProcess();
      }
      catch (Exception ex) {
        Trace.Log(ex);

        response = new ImapResponse(ImapTransactionStatus.BadRequest, ex.Message);

        Transit(null);
      }
    }

    protected void Transit(ProcessTransactionDelegate process)
    {
      currentProcess = process;
    }

    private static int sequenceIdCount = 0;
    private readonly ImapConnection connection;
    private readonly ImapRequest request;
    private ImapResponse response;
    private int sequenceId;
    private ProcessTransactionDelegate currentProcess;
  }
}
