using System;
using System.Collections.Generic;
using System.IO;
using System.Net.Sockets;
using System.Threading;

namespace Smdn.Protocols.Pop3.Protocol {
  // Layers:
  //   Client.PopClient
  //   Client.PopClientAgent
  //   Session.PopSessionAgent
  //   Session.PopSession
  //   Transaction.PopTransactionAgent
  //   Transaction.PopTransaction/Transaction.PopRequest/Transaction.PopResponse
  // * Protocol.PopConnection/Protocol.PopProtocolCommand/Protocol.PopProtocolResponse
  //   (POP3)

  public class PopConnection {
    private class Receiver {
      public Receiver(Stream stream)
      {
        this.stream = stream;
      }

      public byte[] TryReceive()
      {
        lock (receivedData) {
          if (0 < receivedData.Count)
            return receivedData.Dequeue();
        }

        if (readEvent.WaitOne(0, false))
          BeginReceive();

        return null;
      }

      private void BeginReceive()
      {
        stream.BeginRead(buffer, 0, buffer.Length, EndReceive, this);
      }

      private void EndReceive(IAsyncResult ar)
      {
        // TODO: サーバからの切断の考慮
        var read = stream.EndRead(ar);

        if (read <= 0)
          return;

        var data = new byte[read];

        Array.Copy(buffer, data, read);

        lock (receivedData) {
          receivedData.Enqueue(data);
        }

        Trace.Dump(data);

        readEvent.Set();
      }

      private readonly Stream stream;
      private readonly AutoResetEvent readEvent = new AutoResetEvent(true);
      private readonly byte[] buffer = new byte[0x400];
      private readonly Queue<byte[]> receivedData = new Queue<byte[]>();
    }

    /*
    private class Sender {
      // TODO: impl
      public Sender(Stream stream)
      {
        this.stream = stream;
      }

      private readonly Stream stream;
    }
    */

    /*
     * properties
     */

    internal int Id {
      get { return id; }
    }

    public bool Connected {
      get { return connected; }
    }

    public PopServerInfo ServerInfo {
      get { return serverInfo; }
    }

    // for autologout timer
    public DateTime LastSentTime {
      get { return lastSentTime; }
    }

    private PopConnection(PopUri serverUri, int id)
    {
      this.serverInfo = new PopServerInfo(serverUri);
      this.id = id;
    }

    ~PopConnection()
    {
      Disconnect();
    }

    static public PopConnection Connect(PopUri serverUri, PopCreateAuthenticatedStreamCallback createAuthStreamCallback)
    {
      var connection = new PopConnection(serverUri, Interlocked.Increment(ref connectionIdCount));

      if (connectionIdCount == int.MaxValue)
        Interlocked.Exchange(ref connectionIdCount, 0);

      connection.client = new TcpClient();

      if (serverUri.Port == -1)
        connection.client.Connect(serverUri.Host, PopUri.DefaultPort);
      else
        connection.client.Connect(serverUri.Host, serverUri.Port);

      if (createAuthStreamCallback == null)
        connection.stream = connection.client.GetStream();
      else
        connection.stream = createAuthStreamCallback(serverUri, connection.client.GetStream());

      connection.receiver = new Receiver(connection.stream);
      //connection.sender = new Sender(connection.stream);

      // for first receiving
      connection.receiver.TryReceive();

      connection.connected = true;

      Trace.Info("CID:{0} connected to {1} ({2})", connection.Id, connection.client.Client.RemoteEndPoint, serverUri.Host);

      return connection;
    }

    public void Disconnect()
    {
      if (!connected)
        return;

      Trace.Info("CID:{0} disconnecting from {1}", id, client.Client.RemoteEndPoint);

      if (client != null) {
        client.Close();
        client = null;
      }

      Trace.Info("CID:{0} disconnected", id);

      connected = false;
    }

    public PopProtocolCommand CreateCommand(string command, params string[] parameters)
    {
      return new PopProtocolCommand(command, parameters);
    }

    public void SendCommand(PopProtocolCommand command)
    {
      var comm = command.Format();

      Trace.Log(this, comm);

      stream.Write(comm, 0, comm.Length);

      lastSentTime = DateTime.Now;
    }

    public PopProtocolResponse TryReceiveResponse()
    {
      // return parsed response
      if (0 < parsedResponse.Count)
        return parsedResponse.Dequeue();

      // get received data
      byte[] received = receiver.TryReceive();

      if (received == null)
        return null;

      // parse received data
      foreach (var response in parser.Parse(received)) {
        Trace.Log(this, response);

        parsedResponse.Enqueue(response);
      }

      // return parsed response
      if (0 < parsedResponse.Count)
        return parsedResponse.Dequeue();

      return null;
    }

    public override string ToString()
    {
      return string.Format("{{Host={0}, Connected={1}, Id={2}}}",
                           client.Client.RemoteEndPoint,
                           client.Client.Connected,
                           id);
    }

    private static int connectionIdCount = 0;
    private readonly int id = 0;
    private bool connected = false;
    private readonly PopServerInfo serverInfo;
    private DateTime lastSentTime = DateTime.Now;
    private TcpClient client;
    private Stream stream;
    private readonly PopProtocolResponseParser parser = new PopProtocolResponseParser();
    private readonly Queue<PopProtocolResponse> parsedResponse = new Queue<PopProtocolResponse>();
    private Receiver receiver = null;
    //private Sender sender = null;
  }
}
