using System;
using System.IO;
using System.Collections.Generic;
using System.Text;

namespace Smdn.Protocols.Pop3.Protocol {
  public class PopProtocolResponseParser {
    public PopProtocolResponse[] Parse(byte[] responseBytes)
    {
      var responses = new List<PopProtocolResponse>();

      // 受信したものを全てバッファにため込む
      buffer += latin1Encoding.GetString(responseBytes);

      try {
        for (;;) {
          string unparsed;

          var resps = TryParse(buffer, out unparsed);

          // 解析されなかった分をバッファに戻す
          buffer = unparsed;

          if (resps == null) {
            // 解析できなかったので次のレスポンスを待つ(レスポンスの途中の場合)
            break;
          }
          else {
            responses.AddRange(resps);

            // バッファに残っている場合は、継続して解析する
            if (0 < buffer.Length)
              continue;
          }
        }
      }
      catch (Exception ex) {
        Trace.Log(ex);
        Trace.Error("parser error. buffered message was: {0}", buffer);

        // TODO: parser errorの場合の考慮
        buffer = string.Empty; // 捨てる
      }

      return responses.ToArray();
    }

    private IEnumerable<PopProtocolResponse> TryParse(string buffer, out string unparsed)
    {
      if (string.Empty.Equals(buffer)) {
        unparsed = string.Empty;
        return null;
      }

      var crlf = buffer.IndexOf("\r\n");

      if (crlf == -1) {
        unparsed = buffer;
        return null;
      }

      var firstLine = buffer.Substring(0, crlf + 2);
      var trailingLines = buffer.Substring(crlf + 2);

      // There are currently two status
      // indicators: positive ("+OK") and negative ("-ERR").  Servers MUST
      // send the "+OK" and "-ERR" in upper case.
      PopProtocolStatus? status = null;

      if (firstLine.StartsWith("+OK")) {
        status = PopProtocolStatus.Positive;
      }
      else if (firstLine.StartsWith("-ERR")) {
        status = PopProtocolStatus.Negative;
      }

      if (status == null) {
        if (firstLine.StartsWith("+ ")) {
          // PopContinuationRequest
          unparsed = trailingLines;

          // continue_req    ::= "+" SPACE base64 CRLF
          var text = Encoding.ASCII.GetString(Convert.FromBase64String(firstLine.Substring(2)));

          return new[] {new PopContinuationRequest(text, firstLine)};
        }
        else {
          // PopFollowingResponse and PopTerminationResponse
          var responses = new List<PopProtocolResponse>();

          responses.Add(new PopFollowingResponse(firstLine.Substring(0, firstLine.Length - 2), firstLine));

          if (trailingLines.StartsWith(".\r\n")) {
            // PopTerminationResponse
            unparsed = trailingLines.Substring(3);
            responses.Add(new PopTerminationResponse(trailingLines.Substring(0, 3)));
          }
          else {
            unparsed = trailingLines;
          }

          return responses;
        }
      }
      else {
        // PopStatusResponse
        var sepPos = firstLine.IndexOf(" ");
        string message;
        
        if (0 <= sepPos) {
          message = firstLine.Substring(sepPos).Trim();
        }
        else {
          message = string.Empty;
        }

        unparsed = trailingLines;

        return new[] {
          new PopStatusResponse((PopProtocolStatus)status, message, firstLine),
        };
      }
    }

    private static readonly Encoding latin1Encoding = Encoding.GetEncoding("ISO-8859-1");
    private string buffer = string.Empty;
  }
}